@@ -16,8 +16,11 @@ import warnings
 
 try:
     from bson import json_util, SON, BSON, Timestamp
+    from bson import codec_options as bson_codec_options
+    _DEFAULT_CODEC_OPTIONS = bson_codec_options.CodecOptions()
 except ImportError:
-    json_utils = SON = BSON = Timestamp = None
+    bson_codec_options = json_utils = SON = BSON = Timestamp = None
+    _DEFAULT_CODEC_OPTIONS = None
 try:
     import execjs
 except ImportError:
@@ -67,14 +70,14 @@ _KwargOption = collections.namedtuple('KwargOption', ['typename', 'default', 'at
 
 _WITH_OPTIONS_KWARGS = {
     'codec_options': _KwargOption(
-        'bson.codec_options.CodecOptions', None,
+        'bson.codec_options.CodecOptions', _DEFAULT_CODEC_OPTIONS,
         ('document_class', 'tz_aware', 'uuid_representation')),
     'read_preference': _KwargOption(
         'pymongo.read_preference.ReadPreference', _READ_PREFERENCE_PRIMARY,
         ('document', 'mode', 'mongos_mode', 'max_staleness')),
     'write_concern': _KwargOption(
         'pymongo.write_concern.WriteConcern', WriteConcern(),
-        ('acknowledged', 'document', 'is_server_default')),
+        ('acknowledged', 'document')),
     'read_concern': _KwargOption(
         'pymongo.read_concern.ReadConcern', None,
         ('document', 'level', 'ok_for_legacy'))
@@ -263,7 +266,7 @@ class BulkOperationBuilder(object):
 
     def insert(self, doc):
         def exec_insert():
-            self.collection.insert(doc)
+            self.collection.insert_one(doc)
             return {'nInserted': 1}
         self.executors.append(exec_insert)
 
@@ -359,12 +362,16 @@ class BulkOperationBuilder(object):
 
 class Collection(object):
 
-    def __init__(self, database, name, _db_store, write_concern=None, read_preference=None):
+    def __init__(
+            self, database, name, _db_store, write_concern=None, read_preference=None,
+            codec_options=None):
         self.database = database
         self._name = name
         self._db_store = _db_store
         self._write_concern = write_concern or WriteConcern()
         self._read_preference = read_preference or _READ_PREFERENCE_PRIMARY
+        self._codec_options = codec_options
+        self._read_concern = None
 
     def __repr__(self):
         return "Collection({0}, '{1}')".format(self.database, self.name)
@@ -395,6 +402,14 @@ class Collection(object):
     def read_preference(self):
         return self._read_preference
 
+    @property
+    def codec_options(self):
+        if not bson_codec_options:
+            raise NotImplementedError(
+                'The codec options are not implemented in mongomock alone, you need to import '
+                'the pymongo library as well.')
+        return self._codec_options
+
     def initialize_unordered_bulk_op(self):
         return BulkOperationBuilder(self, ordered=False)
 
@@ -418,30 +433,41 @@ class Collection(object):
         documents = list(documents)
         for document in documents:
             validate_is_mutable_mapping('document', document)
-        return InsertManyResult(self._insert(documents, session), acknowledged=True)
+        return InsertManyResult(
+            self._insert(documents, session, ordered=ordered),
+            acknowledged=True)
 
     @property
     def _store(self):
         return self._db_store[self._name]
 
-    def _insert(self, data, session=None):
+    def _insert(self, data, session=None, ordered=True):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
         if not isinstance(data, Mapping):
             results = []
+            write_errors = []
+            num_inserted = 0
             for index, item in enumerate(data):
                 try:
                     results.append(self._insert(item))
                 except WriteError as error:
-                    raise BulkWriteError({
-                        'writeErrors': [{
-                            'index': index,
-                            'code': error.code,
-                            'errmsg': str(error),
-                            'op': item,
-                        }],
-                        'nInserted': index,
+                    write_errors.append({
+                        'index': index,
+                        'code': error.code,
+                        'errmsg': str(error),
+                        'op': item,
                     })
+                    if ordered:
+                        break
+                    else:
+                        continue
+                num_inserted += 1
+            if write_errors:
+                raise BulkWriteError({
+                    'writeErrors': write_errors,
+                    'nInserted': num_inserted,
+                })
             return results
 
         # Like pymongo, we should fill the _id in the inserted dict (odd behavior,
@@ -670,7 +696,15 @@ class Collection(object):
                             arr_copy = copy.deepcopy(arr)
                             if isinstance(value, dict):
                                 for obj in arr_copy:
-                                    if filter_applies(value, obj):
+                                    try:
+                                        is_matching = filter_applies(value, obj)
+                                    except OperationFailure:
+                                        is_matching = False
+                                    if is_matching:
+                                        arr.remove(obj)
+                                        continue
+
+                                    if filter_applies({field: value}, {field: obj}):
                                         arr.remove(obj)
                             else:
                                 for obj in arr_copy:
@@ -929,6 +963,9 @@ class Collection(object):
                     if sort_direction < 0:
                         dataset = iter(reversed(list(dataset)))
                     continue
+                if sort_key.startswith('$'):
+                    raise NotImplementedError(
+                        'Sorting by {} is not implemented in mongomock yet'.format(sort_key))
                 dataset = iter(sorted(
                     dataset, key=lambda x: resolve_sort_key(sort_key, x),
                     reverse=sort_direction < 0))
@@ -1332,7 +1369,7 @@ class Collection(object):
         is_unique = kwargs.pop('unique', False)
         is_sparse = kwargs.pop('sparse', False)
 
-        index_string = helpers.gen_index_name(index_list)
+        index_name = kwargs.pop('name', helpers.gen_index_name(index_list))
         index_dict = {'key': index_list}
         if is_sparse:
             index_dict['sparse'] = True
@@ -1341,6 +1378,7 @@ class Collection(object):
         if is_unique:
             index_dict['unique'] = True
             indexed = set()
+            indexed_list = []
             for doc in self._store.documents:
                 index = []
                 for key, unused_order in index_list:
@@ -1353,13 +1391,19 @@ class Collection(object):
                 if is_sparse and not index:
                     continue
                 index = tuple(index)
-                if index in indexed:
-                    raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
-                indexed.add(index)
+                try:
+                    if index in indexed:
+                        raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
+                    indexed.add(index)
+                except TypeError:
+                    # index is not hashable.
+                    if index in indexed_list:
+                        raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
+                    indexed_list.append(index)
 
-        self._store.indexes[index_string] = index_dict
+        self._store.indexes[index_name] = index_dict
 
-        return index_string
+        return index_name
 
     def create_indexes(self, indexes, session=None):
         for index in indexes:
@@ -1371,7 +1415,8 @@ class Collection(object):
                 index.document['key'].items(),
                 session=session,
                 unique=index.document.get('unique', False),
-                sparse=index.document.get('sparse', False))
+                sparse=index.document.get('sparse', False),
+                name=index.document.get('name'))
             for index in indexes
         ]
 
@@ -1594,30 +1639,31 @@ class Collection(object):
         in_collection = [doc for doc in self.find()]
         return aggregate.process_pipeline(in_collection, self.database, pipeline, session)
 
-    def with_options(self, **kwargs):
-        forbidden_kwargs = set(kwargs.keys()) - set(_WITH_OPTIONS_KWARGS)
-        if forbidden_kwargs:
-            raise TypeError(
-                "with_options() got an unexpected keyword argument '%s'" % forbidden_kwargs.pop())
-
-        for key, value in iteritems(kwargs):
-            if value is None:
+    def with_options(
+            self, codec_options=None, read_preference=None, write_concern=None, read_concern=None):
+        has_changes = False
+        for key, options in iteritems(_WITH_OPTIONS_KWARGS):
+            value = locals()[key]
+            if value is None or value == getattr(self, '_' + key):
                 continue
-            options = _WITH_OPTIONS_KWARGS[key]
+            has_changes = True
             for attr in options.attrs:
                 if not hasattr(value, attr):
                     raise TypeError(
                         '{} must be an instance of {}'.format(key, options.typename))
-            if key not in ('read_preference', 'write_concern') and options.default != value:
+            if key in ('codec_options', 'read_concern') and options.default != value:
+                # TODO(pascal): Support change of tz_aware in codec_options.
                 raise NotImplementedError(
                     '%s is a valid parameter for with_options but it is currently not implemented '
                     'in Mongomock' % key)
 
-        write_concern = kwargs.get('write_concern', self.write_concern)
-        read_preference = kwargs.get('read_preference', self.read_preference)
+        if not has_changes:
+            return self
+
         return Collection(
-            self.database, self.name, write_concern=write_concern, read_preference=read_preference,
-            _db_store=self._db_store)
+            self.database, self.name, write_concern=write_concern or self._write_concern,
+            read_preference=read_preference or self._read_preference,
+            codec_options=codec_options or self._codec_options, _db_store=self._db_store)
 
     def rename(self, new_name, session=None, **kwargs):
         if session:
@@ -1663,6 +1709,7 @@ class Cursor(object):
     def _compute_results(self, with_limit_and_skip=False):
         # Recompute the result only if the query has changed
         if not self._results or self._factory_last_generated_results != self._factory:
+            # TODO(pascal): Check collection's codec_options instead of reaching to the client.
             if self.collection.database.client._tz_aware:
                 results = [helpers.make_datetime_timezone_aware_in_document(x)
                            for x in self._factory()]
@@ -1701,12 +1748,18 @@ class Cursor(object):
         self._emitted = 0
 
     def sort(self, key_or_list, direction=None):
-        self._sort = helpers.create_index_list(key_or_list, direction)
+        sort = helpers.create_index_list(key_or_list, direction)
+        if not sort:
+            raise ValueError('key_or_list must not be the empty list')
+        self._sort = sort
         self._factory = functools.partial(
             self.collection._get_dataset, self._spec, self._sort, self._projection, dict)
         return self
 
     def count(self, with_limit_and_skip=False):
+        warnings.warn(
+            'count is deprecated. Use Collection.count_documents instead.',
+            DeprecationWarning, stacklevel=2)
         results = self._compute_results(with_limit_and_skip)
         return len(results)
 
